from remotemanager.storage.sendablemixin import SendableMixin


class placeholder_option(SendableMixin):
    """
    Stub class to sit in place of an option within a computer.
    """

    def __init__(self, mode):
        self.mode = mode

    def __hash__(self):
        return hash(self.mode)

    def __bool__(self):
        """
        Make the objects "falsy" this enables rapid defaulting:

        `value = option or default`
        """
        return False


required = placeholder_option('required')
optional = placeholder_option('optional')



"""
Handles file transfer via `cp`
"""
import logging

from remotemanager.transport.transport import Transport


class cp(Transport):
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._logger.info('created new cp transport')

        self._cmd = 'mkdir -p {secondary} ; ' \
                    'cp -r --preserve {primary}{files} {secondary}'

    @property
    def cmd(self):
        base = self._cmd.format(primary='{primary}',
                                files='{files}',
                                secondary='{secondary}')
        self._logger.debug(f'returning semi-formatted cmd: "{base}"')
        return base



import gc
import os

from remotemanager.storage.sendablemixin import SendableMixin


class Dependency(SendableMixin):

    _valid_modes = ['one2one']

    def __init__(self,
                 parent,
                 child,
                 mode):

        if parent == child:
            raise ValueError('parent and child are the same!')

        self._uuids = {parent.uuid: 'parent',
                       child.uuid: 'child'}
        self._parent = parent
        self._child = child

        self._mode = None
        self.mode = mode

        self._network = {}

    @property
    def child(self):
        if not hasattr(self, '_child'):
            c_uuid = [k for k, v in self._uuids.items()
                      if v == 'child'][0]
            self._child = _dataset_from_uuid(c_uuid)
        return self._child

    @property
    def parent(self):
        if not hasattr(self, '_parent'):
            p_uuid = [k for k, v in self._uuids.items()
                      if v == 'parent'][0]
            self._parent = _dataset_from_uuid(p_uuid)
        return self._parent

    def is_child(self,
                 uuid: str) -> bool:
        """
        Returns True if uuid is a child Dataset
        """
        return self._uuids[uuid] == 'child'

    def is_parent(self,
                  uuid: str) -> bool:
        """
        Returns True if uuid is a parent Dataset
        """
        return self._uuids[uuid] == 'parent'

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode not in Dependency._valid_modes:
            raise ValueError(f'mode {mode} must be one of '
                             f'{Dependency._valid_modes}')

        self._mode = mode

    def append_runs(self, *args, **kwargs):
        key = self.next_dependency_key

        # must append children first to access their scripts within the parent

        func_args = kwargs.pop('args')

        self.child.append_run(*args,
                              dependency_call=True,
                              dependency_key=key,
                              **kwargs)
        self.parent.append_run(*args,
                               arguments=func_args,
                               dependency_call=True,
                               dependency_key=key,
                               **kwargs)

        # create any dependency info
        prv_resultfile = self.get_relative_parent_remote(
            self.parent.runners[-1].resultfile)
        self.child.runners[-1]._dependency_info['child'] = True,
        self.child.runners[-1]._dependency_info['parent_import'] = f'repo.loaded = repo.load("{prv_resultfile}")'

        submit_child = ['import subprocess',
                        f'cmd = "{self.parent.submitter} {self.child.runners[-1].scriptfile}"',
                        'subprocess.Popen(cmd, shell=True, executable="/bin/bash")']
        self.parent.runners[-1]._dependency_info['parent'] = True,
        self.parent.runners[-1]._dependency_info['child_submit'] = '\n'.join(submit_child)

        self.network[key] = (self.parent.runners[-1],
                             self.child.runners[-1])

    def get_relative_parent_remote(self, file):
        if self.child.remote_dir == self.parent.remote_dir:
            return file
        return os.path.join('..', self.parent.remote_dir, file)

    @property
    def next_dependency_key(self):
        return f'{len(self._network)+1}'.rjust(5, '0')

    @property
    def network(self):
        return self._network

    def run(self,
            force,
            dry_run,
            **run_args):
        """
        Submit the parent and child datasets.

        Arguments are the same as the Dataset.run() function
        """

        self.parent._run(force,
                         dry_run,
                         **run_args)
        self.child._run()


def _dataset_from_uuid(uuid):

    for obj in gc.get_objects():
        if obj.__class__.__name__ == 'Dataset':
            if obj.uuid == uuid:
                return obj

    raise ValueError(
        f'could not find Dataset in memory with UUID {uuid}')



"""
Baseclass for any file transfer
"""
import os.path

from remotemanager.connection.cmd import CMD
from remotemanager.connection.url import URL
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.logging.utils import format_iterable
from remotemanager.utils import ensure_list, ensure_dir
from remotemanager.logging import LoggingMixin
from remotemanager.utils.flags import Flags


class Transport(SendableMixin, LoggingMixin):
    """
    Baseclass for file transfer

    Args:
        url (URL):
            url to extract remote address from
    """

    def __init__(self,
                 url: URL = None,
                 *args,
                 **kwargs):

        self._remote_address = None
        self.set_remote(url)

        self._flags = Flags()
        self._transfers = {}

    def queue_for_push(self,
                       files: list,
                       local: str = None,
                       remote: str = None):
        """
        Queue file(s) for sending (pushing)

        Args:
            files (list[str], str):
                list of files (or file) to add to push queue
            local (str):
                local/origin folder for the file(s)
            remote (str):
                remote/destination folder for the file(s)
        Returns:
            None
        """
        self._logger.info(f'adding to PUSH queue')
        self.add_transfer(files, local, remote, 'push')

    def queue_for_pull(self,
                       files: list,
                       local: str = None,
                       remote: str = None):
        """
        Queue file(s) for retrieving (pulling)

        Args:
            files (list[str], str):
                list of files (or file) to add to pull queue
            local (str):
                local/destination folder for the file(s)
            remote (str):
                remote/origin folder for the file(s)
        Returns:
            None
        """
        self._logger.info(f'adding to PULL queue')
        self.add_transfer(files, remote, local, 'pull')

    def add_transfer(self,
                     files: list,
                     origin: str,
                     target: str,
                     mode: str):
        """
        Create a transfer to be executed. The ordering of the origin/target
        files should be considered as this transport instance being a
        "tunnel" between wherever it is executed (origin), and the destination
        (target)

        Args:
            files (list[str], str):
                list of files (or file) to add to pull queue
            origin (str):
                origin folder for the file(s)
            target (str):
                target folder for the file(s)
            mode (str: "push" or "pull"):
                transfer mode. Chooses where the remote address is placed
        Returns:
            None
        """
        modes = ('push', 'pull')
        if mode.lower() not in modes:
            raise ValueError(f'mode must be one of {modes}')

        if origin is None:
            origin = '.'
        if target is None:
            target = '.'

        # ensure dir-type
        origin = os.path.join(origin, '')
        target = os.path.join(target, '')

        if mode == 'push':
            pair = f'{origin}>{self._add_address(target)}'
        else:
            pair = f'{self._add_address(origin)}>{target}'

        files = [os.path.split(f)[1] for f in ensure_list(files)]

        self._logger.info(f'adding transfer: {split_pair(pair)[0]} '
                          f'-> {split_pair(pair)[1]}')
        self._logger.info(f'for files {files}')

        if pair in self._transfers:
            self._transfers[pair] = self._transfers[pair].union(set(files))
        else:
            self._transfers[pair] = set(files)

    def _add_address(self, dir: str) -> str:
        """
        Adds the remote address to the dir `dir` if it exists

        Args:
            dir (str):
                remote dir to have address appended

        Returns:
            (str) dir
        """
        if self.address is None:
            return dir
        return f'{self.address}:{dir}'

    @staticmethod
    def _format_for_cmd(inp: list) -> str:
        """
        Formats a list into a bash expandable command with brace expansion

        Args:
            inp (list):
                list of items to compress

        Returns (str):
            formatted cmd
        """

        if isinstance(inp, str):
            raise ValueError('files is stringtype, '
                             'was a transfer forced into the queue?')
        if len(inp) > 1:
            return '{' + ','.join(inp) + '}'
        return inp[0]

    @property
    def transfers(self) -> dict:
        """
        Return the current transfer dict

        Returns (dict):
            {paths: files} transfer dict
        """
        return {k: sorted(list(v)) for k, v in self._transfers.items()}

    def print_transfers(self):
        """
        Print a formatted version of the current queued transfers

        Returns:
            None
        """
        i = 0
        for pair, files in self.transfers.items():
            i += 1
            print(f'transfer {i}:'
                  f'\norigin: {split_pair(pair)[0]}'
                  f'\ntarget: {split_pair(pair)[1]}')
            j = 0
            for file in files:
                j += 1
                print(f'\t({j}/{len(files)}) {file}')

    @property
    def address(self):
        """
        return the remote address

        Returns (str):
            the remote address
        """
        return self._remote_address

    @address.setter
    def address(self, remote_address):
        """
        set the remote address

        Returns:
            None
        """
        self._remote_address = remote_address

    def set_remote(self,
                   url: URL = None):
        """
        set the remote address with a URL object

        Returns:
            None
        """
        if url is None:
            self._remote_address = None
        elif url.is_local:
            self._remote_address = None
        else:
            self._remote_address = url.userhost

    @property
    def flags(self):
        return self._flags

    @flags.setter
    def flags(self, new):
        self._flags = Flags(str(new))

    @property
    def cmd(self):
        """
        Returns a semi formatted command for issuing transfers. It is left to
        the developer to implement this method when adding more transport
        classes

        Returns (str):
            the semi-formatted command for issuing a transfer
        """
        raise NotImplementedError

    def transfer(self,
                 dry_run: bool = False):
        """
        Perform the actual transfer

        Args:
            dry_run (bool):
                do not perform command, just return the command(s) to be
                executed

        Returns (str, None):
            the dry run string, or None
        """

        self._logger.info(f'executing a transfer')

        commands = []
        for pair, files in self.transfers.items():

            filestr = self._format_for_cmd(files)

            primary = ensure_dir(os.path.split(split_pair(pair)[0])[0])
            secondary = ensure_dir(os.path.split(split_pair(pair)[1])[0])

            base_cmd = self.cmd.format(primary=primary,
                                       secondary=secondary,
                                       files=filestr)

            commands.append(base_cmd)

        if dry_run:
            return commands

        for cmd in commands:
            CMD(cmd).exec()
            # wipe the transfer queue
            self.wipe_transfers()

    def wipe_transfers(self):
        self._logger.info('wiping transfers')
        self._transfers = {}


def split_pair(pair: str) -> list:
    """
    Convert a "dir>dir" string into list format

    Args:
        pair (tuple):
            (dir, dir) tuple to be split

    Returns (list):
        [dir, dir]

    """
    return pair.split('>')
